# -*- coding: utf-8 -*-
# Python Smart Cloud Communication Library 0.1-dev
#


"""
TODO
* Nimekiri käskudest
* stukruuri korrastamine
* Korralik dokumentatsioon
* Connect to asjade sulgemine
    või noh üldse kõige sulgemine puhtalt


Random info
* kõigi seadmete otsimiseks tuleb saata käsk 0x000e aadressile 255, 255
  käsul parameeterid pole
* seadmelt temperatuuri küsimiseks on käsk 0xe3e7, parameetriks on [1]
  vastuseks tuleb operate code 0xe3e8 kahe parameetriga, teine parameter
  on temperatuur esimene parameeter vb on kanali number



    How to use
faili lõpus on näide


"""

# import stuff
import struct  # for encoding/decoding smartcloud packets
import socket  # sending smartcloud packets
#import SocketServer #reciving smartcloud packets
import threading
import Queue  # fifo for storing recived smartcloud packets
import time


class Timeout(Exception):
    """ this exception is triggerd when packet does not arive in given time """
    pass


def crc_calc(buff):
    """Calculates smart-bus packet crc code"""
    # i have no idea how it works
    crc = 0
    for x in buff:
        i = ord(x)
        crc = crc ^ (i << 8)
        for asdf in range(1, 9):
            if (crc & 0x8000) > 0:
                crc = (crc << 1) ^ 0x1021
            else:
                crc = crc << 1
        crc = crc & 0xFFFF
    return (crc >> 8, crc & 0xFF)


class SmartCloudSend():
    """
    Encodes and sends out smartcloud packets

    little trick
    a = ".".join([1, 2, 3, 4])
        a = "1.2.3.4"

    b = a.split(".")
        b = [1, 2, 3, 4]
    """
    def __init__(self,
                MyIP=[192, 168, 1, 10], #doesnt do realy anything
                #MyBroadcastIP = [192,168,1,255],
                #MyBroadcastIP=[192, 168, 5, 255],
                MyBroadcastIP="<broadcast>",
                MySubnet=1,
                MyID=250,
                UDPPort=6000,
                MyType=[255, 253]):
        """ docstring """
        self.MyIP = MyIP
        self.MySubnet = MySubnet
        self.MyID = MyID
        self.MyType = MyType
        self.MyBroadcastIP = MyBroadcastIP
        self.UDPPort = UDPPort

        # setup udp stuff
        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.udp_socket.bind(('', 0))
        self.udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    def send_udp(self, Data, UDPTo=None, UDPPort=None):
        if UDPTo is None: UDPTo = self.MyBroadcastIP  # because python is
        if UDPPort is None: UDPPort = self.UDPPort    # weird
        """
        Sends UDP packages

        """
        #UDPTo = ".".join(UDPTo)
        tmp = UDPTo
        if type(UDPTo) is list:  # ip address list to string
            UDPTo = ""
            UDPTo += str(tmp[0])
            UDPTo += "."
            UDPTo += str(tmp[1])
            UDPTo += "."
            UDPTo += str(tmp[2])
            UDPTo += "."
            UDPTo += str(tmp[3])

        self.udp_socket.sendto(Data, (UDPTo, UDPPort))

    def make_package(self, Command, CommandParam,
                    ToDevID, ToSubnet, protocol_id="SMARTCLOUD"):
        """
        This function generates Smart Cloud UDP package witch is ready to be
        send

        """
        package = []
        package.extend(self.MyIP)  # insert our ip
        package.extend(protocol_id)  # our smartcloud string
        package.extend([0xAA, 0xAA])  # package start
        package.append(9 + len(CommandParam) + 2)  # (header+command)+param+crc
        package.append(self.MySubnet)
        package.append(self.MyID)
        package.extend(self.MyType)
        package.append(Command / 256)
        package.append(Command % 256)
        package.append(ToSubnet)
        package.append(ToDevID)
        package.extend(CommandParam)  # add comand parameters
        crc = crc_calc(struct.pack('{0}B'.format(len(package[16:])), *package[16:]))
        package.extend(crc)  # add crc code

        #turn package to binnary format
        package_bin = struct.pack('4B 10c 2B {0}B'.format(len(package[16:])), *package)
        return package_bin

    def custom(self, command=0x0000, param=[], ToSubnet = 1, ToDevID = 1, protocol_id="SMARTCLOUD"):
        """
        Send customized smartcloud pakcages"""
        #print "sending:cmd:{0} prm:{1} tosub:{2} todev:{3}".format(hex(command),
        #                                                           str(param),
        #                                                           str(ToSubnet),
        #                                                           str(ToDevID))
        packet = self.make_package(command, param, ToDevID, ToSubnet, protocol_id)
        self.send_udp(packet)
        return "ok"

    def dimmer(self, Channel = 1, Level = 100, ToSubnet = 1, ToDevID = 16):
        """
        Channel 255 turns all lights on
        Level = 0 - 100
        """
        self.custom(command=0x0031, param=[int(Channel), int(Level), 0, 0, 0],
                    ToSubnet = ToSubnet, ToDevID = ToDevID)
        return "ok"

    def relay(self, Channel = 1, Level = 100, ToSubnet = 1, ToDevID = 29):
        """
        well becasuse relay uses exacly the same code as light dimmer
        level 100 = on
        level 0 = off
        """
        return self.dimmer(Channel, Level, ToSubnet, ToDevID)

    def dev_search(self, ToSubnet = 255, ToDevID = 255):
        """
        Send out device search packet
        """
        return self.custom(command=0x000e, param=[], ToSubnet = ToSubnet, ToDevID = ToDevID)

    def send(self,*largs, **kwargs):
        "Send customized smartcloud packages"
        self.custom(*largs, **kwargs)

class SmartCloudRecive():
    """
    Listens and decodes smartcloud packets
    """

    def __init__(self,
                ListenIP="",
                ListenPort=6000,):

        self.ListenIP = ListenIP
        self.ListenPort = ListenPort
        self.qPackets = Queue.Queue()
        self.capture_broken = False
        # create a huge list with all possible smartcloud addresses
        self._dev_listeners_ = [[[]for x in range(255)] for x in range(255)]

    def register_dev_listener(self, function, subnet, id):
        """run function when packet with corrent subnet and id is captured
        Usage:
            register_dev_listener(self, function, subnet, id)"""
        self._dev_listeners_[subnet][id].append(function)

    def remove_dev_listener(self,function, subnet, id):
        """Remove device lisntener funcion from list"""
        l = self._dev_listeners_[subnet][id]
        del l[l.index(function)]

    def _run_listener_(self, packet):
        "runs registred funtions"
        # dev listener
        for funct in self._dev_listeners_[packet["from_device_subnet"]][packet["from_device_id"]]:
            try:
                funct(packet)  # run funtion
            except:
                print "broken function:", funct

    def get(self):
        return self.qPackets.get()

    def recive(self, mis_see_on):  # thread
        """
        Starts lisening for smartcloud packets.
        """
        sock = socket.socket(socket.AF_INET,  # Internet
                            socket.SOCK_DGRAM)  # UDP
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((self.ListenIP, self.ListenPort))
        qPackets = self.qPackets
        capture_broken = self.capture_broken
        """
        SmartCloud pakettide dekodeerimine ja fifo'se lisamine
        """
        while True:
            #try:
            packet_raw, addr = sock.recvfrom(254)
            packet = {}
            #print packet_raw
            packet_un = struct.unpack("4B 10c 2B {0}B BB".format(len(packet_raw)-18), packet_raw)

            packet.update(dict(source_ip=packet_un[0:4])) # paketi teele pannud smartbus ip mooduli lan võrgu aadress
            packet.update(dict(smartcloud=packet_un[4:14])) # paketi tüübi tuvastamiseks, meil peab see andma vastuseks "SMARTCLOUD"
            packet.update(dict(packet_start=packet_un[14:16])) # Smartbus paketi alguse märge
            packet.update(dict(packet_size=packet_un[16])) # kui pikk on smartbus pakett
            packet.update(dict(from_device_subnet=packet_un[17])) # paketi teele pannud smartbus seadme segment
            packet.update(dict(from_device_id=packet_un[18])) # paketi teele pannud smartbus seadme id
            packet.update(dict(from_device_type=packet_un[18:20]))
            packet.update(dict(operate_code=packet_un[21] * 256 + packet_un[22]))
            packet.update(dict(to_dev_id=packet_un[24])) #millisesse seadmesse pakett teel on
            packet.update(dict(to_subnet=packet_un[23]))
            packet.update(dict(data=packet_un[25:-2])) # pakett ise mis on muutuva suurusega
            packet.update(dict(CRC=packet_un[-2:])) # kontroll kood
            packet.update(dict(CRC_OK=packet["CRC"] == crc_calc(packet_raw[16:-2])))
            packet.update(dict(timestamp=time.time()))

            #check if crc is correct
            if packet["CRC_OK"]:
                qPackets.put(packet)
                self._run_listener_(packet)
            elif capture_broken:
                qPackets.put(packet)
                self._run_listener_(packet)
            else:
                print "Broken package"
            #except:
            #    print "perse, decode pani pange"

    def start(self):
        t = threading.Thread(name='sbus recive',
                            target=self.recive,
                            args=(self,))
        t.setDaemon(True)
        t.start()
        return "OK"

class ConnectTo(object):
    """Connect to device to do easily two-way communication
        Usage:
            con = ConnectTo(subnet, id,
                        SmartCloudSend instance, SmartCloudRecive instance,
                        [protocol="SMARTCLOUD"])
            con.write(cmd, data)
            recive = con.read([time_out])
            print recive # dict("cmd", "data", "timestamp")
    """

    def __init__(self, subnet, id,
                        scs_obj, scr_obj,
                        protocol="SMARTCLOUD"):
        """ConnectTo(subnet, id,
                    scs_obj=SmartCloudSend, scr_obj=SmartCloudRecive,
                    protocol="SMARTCLOUD" """
        self.subnet = subnet
        self.id = id
        self.scs_obj = scs_obj
        self.scr_obj = scr_obj
        self.protocol = protocol
        self._qPackets_ = Queue.Queue()
        self._last_cmd_ = None
        # register listener
        scr_obj.register_dev_listener(self._got_package_, int(subnet), int(id))

    def close(self):
        self.scr_obj.remove_dev_listener(self._got_package_, int(self.subnet), int(self.id))

    def _got_package_(self, packet):
        new_packet = {}
        new_packet["cmd"] = int(packet["operate_code"])
        new_packet["data"] = list(packet["data"])
        #new_packet["timestamp"] = int(packet["timestamp"])  # mõtetu
        self._qPackets_.put(new_packet)
        #print "DEBUG, got packet", new_packet

    def write(self, cmd, data=[]):
        """Send commands to device
        Usage:
            write(cmd, data)"""
        #time.sleep(0.5)
        self.scs_obj.custom(command=cmd, param=data,
                ToSubnet=self.subnet, ToDevID=self.id,
                protocol_id=self.protocol)
        self._last_cmd_ = cmd

    def read(self,cmd=-1, timeout=10):
        """read back data
            Usage:
                read(cmd, timeout)
                if cmd == -1 then read package last write cmd +1
                    (if has no previus cmd then return any packet)
                if cmd = None then return first available package
                timeout is in seconds"""
        while True:
            #print "DEBUG, read, loop start"
            try:
                packet = self._qPackets_.get(True, timeout) #wait for a packet
            except Queue.Empty:
                raise Timeout()
            if cmd == -1:
                if self._last_cmd_ != None:
                    if packet["cmd"] == self._last_cmd_ + 1:
                        #print "DEBUG packet with cmd", self._last_cmd_ +1
                        return packet
                else:
                    return packet
            elif cmd == None:
                return packet
            elif packet["cmd"] == cmd:
                return packet




if __name__ == '__main__':
    from time import sleep
    #from pprint import pprint
    import sys

    def pprint(packet):
        #print "from id:", packet["from_device_id"], "subnet:", packet["from_device_subnet"]
#        print "from device type:", packet["from_device_type"]
#        print "operate code:", hex(packet["operate_code"]), "(dec:",packet["operate_code"],")"
        #print "to id:", packet["to_dev_id"], "subnet:", packet["to_subnet"]
        #print "packet data:", packet["data"]
        #print

        packet["from_device_type"] = packet["from_device_type"][0]*256+packet["from_device_type"][1]
        packet["data_ascii"] = "".join([chr(x) if x > 31 else "?" for x in packet["data"]])
        if packet["operate_code"] == 0x2c0:
            print "[Asking for Date Time]"
        elif packet["operate_code"] == 0x2c1:
            print "[Got Date Time]"
        elif packet["operate_code"] == 0xdd10:
            print "[Asking packet count from impuse counter]"

        print """from sub={0}, id={1}, opcode={2}""".format(str(packet["from_device_subnet"]),
                                                    str(packet["from_device_id"]),
                                                    str(hex(packet["operate_code"])))
        print """to sub={0}, id={1}""".format(str(packet["to_subnet"]), str(packet["to_dev_id"]))
        print """data={0}""".format(str(packet["data"]))
            # print """Z1={0}, Z2={1}, Z3={2}, Z4={3}""".format(str(packet["data"][5]),
            #                                                 str(packet["data"][6]),
            #                                                 str(packet["data"][7]),
            #                                                 str(packet["data"][8]))
        print



    print  "Defining stuff"
    #m = SmartCloudRecive(ListenIP = "255.255.255.255")
    m = SmartCloudRecive("192.168.10.255")
    s = SmartCloudSend(MyBroadcastIP="<broadcast>")

    print "Starting to listen for packets"
    m.start()

    args = ""
    try:
        args = sys.argv[1]
    except:
        pass

    if args == "-s":
        #send some test packets
        while True:
            print "sending packet\n"
#            s.custom(command=0x1946, param=[0, 1, 1, 23, 20, 20, 20], ToSubnet = 35, ToDevID = 4)
#            s.custom(command=0x1944, param=[], ToSubnet = 35, ToDevID = 4)
            s.custom(command=0xe3e7, param=[1], ToSubnet=35, ToDevID=10)
            sleep(1)
            while not m.qPackets.empty():
                pprint(m.qPackets.get())
            sleep(1)
    elif args == "-f":
        dev_id = int(sys.argv[2])
        print "filter mode"
        print "showing only packets from device", dev_id
        print "listening for smartcloud packets"
        while True:
            if not m.qPackets.empty():
                packet = m.qPackets.get()
                if packet["from_device_id"] == dev_id:
                    pprint(packet)
#                else:
#                    print "got packet from device", packet["from_device_id"]
            else:
                sleep(0.01)
    else:
        print "Kuulan pakette"
        while True:
            if not m.qPackets.empty():
                #print
                pprint(m.qPackets.get())
            else:
                sleep(0.01)

